
/* 
   Copyright 1997, Kevin J. Lang 

   chunky_decode is free software, licensed under the 
   GNU Public License V2.

   See the file COPYING for details.
*/

/*  
    how to compile: 
    gcc -O2 -Wall -g -o chunky_decode chunky_decode.c  

 */


#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#define dotimes(var,lim) for (var=0;var<(lim);var++) 
#define MAX(a,b) (((a)>(b)) ? (a) : (b))
#define MIN(a,b) (((a)<(b)) ? (a) : (b))
#define ABS(zz) ((zz)<0 ? -1*(zz) : (zz))


#define CHUNK_SIZE 5000

void my_chuffman_decode (unsigned char * inbuf,
			 unsigned char * outbuf, int outlen,
			 short * chuftab_offsets,
			 unsigned char * chuftab,
			 unsigned char * unmap)
{
  int bit_counter;
  unsigned char * inptr = inbuf;
  unsigned char bits, bit;
  unsigned char * outptr = outbuf;
  unsigned char * outend = outbuf + outlen;
  short curstate_idx = 0;
  unsigned char curstate;
  unsigned char * huftab = chuftab;
  unsigned char decoded;

  while (1) {
    bits = *inptr++;
    bit_counter = 0;
    while (bit_counter < 8) {
      curstate = huftab [curstate_idx];
      if (curstate >= 170) {
	decoded = curstate - 170;
	*outptr++ = unmap [decoded];
	huftab = chuftab + chuftab_offsets [decoded];
	if (outptr == outend) return;
	curstate_idx = 0;
	continue;
      }
      bit = bits & 1;
      bits = bits >> 1;
      bit_counter += 1;
      if (bit) curstate_idx = curstate;
      else curstate_idx = curstate_idx + 1;
    }
  }
}



/*************************************************/

int n_uniq = 0;

unsigned char * unmap_table;

short * decoder_offsets;

unsigned char * decoder_table;


void read_code_file (char * name)
{
  FILE * input_file;
  int i;
  int len;
  int itmp;

  input_file = fopen (name, "r");
  assert(input_file != NULL);

  fscanf(input_file, "%d", &n_uniq);
  unmap_table = (unsigned char *) malloc (n_uniq * sizeof(unsigned char));

  dotimes (i, n_uniq) {
    fscanf(input_file, "%d", &itmp);
    unmap_table [i] = itmp;
  }

  decoder_offsets = (short *) malloc (n_uniq * sizeof(short));

  dotimes (i, n_uniq) {
    fscanf(input_file, "%d", &itmp);
    decoder_offsets [i] = itmp;
  }

  fscanf(input_file, "%d", &len);
  decoder_table = (unsigned char *) malloc (len * sizeof(unsigned char));

  dotimes (i, len) {
    fscanf(input_file, "%d", &itmp);
    decoder_table [i] = itmp;
  }

  fclose (input_file);

}


int n_chunks;

unsigned char inbuf  [CHUNK_SIZE];
unsigned char outbuf [CHUNK_SIZE];

int bpos;

void get_word (unsigned char * buf)
{
  int c;
  unsigned char * bpt;

  c = outbuf[bpos];
  bpos += 1;
  assert (c <= 15);
  bpt = buf + c;

  if (outbuf[bpos] == 25) {
    bpos += 1;
    *bpt += 1;
    bpt += 1;
  }

  while ((c = outbuf[bpos]) > 15) {
    *bpt++ = c;
    bpos += 1;
  }

  *bpt = '\0';
}






int main (int argc, char **argv)
{
  int i, j;
  FILE * infile;
  unsigned char wordbuf [100];

  if (argc != 4) {
    fprintf(stderr, "usage: %s infile code-file n-chunks > out-file\n",
	    argv[0]);
    exit(-1);
  }

  infile = fopen(argv[1], "r");

  read_code_file (argv[2]);

  n_chunks = atoi (argv[3]);

  dotimes (i, n_chunks) {

    short inlen, outlen, nwords;

    fread (&inlen, sizeof(short), 1, infile);
    fread (&outlen, sizeof(short), 1, infile);
    fread (&nwords, sizeof(short), 1, infile);
    fread (inbuf, 1, inlen, infile);

    my_chuffman_decode (inbuf, outbuf, outlen,
			decoder_offsets, decoder_table,
			unmap_table);
    bpos = 0;
    dotimes (j, nwords) {
      get_word(wordbuf);
      printf("%s\n", wordbuf);
    }
  }

  fclose(infile);

  return(0);
}
