
/* 
   Copyright 1997, Kevin J. Lang 

   chunky_encode is free software, licensed under the 
   GNU Public License V2.

   See the file COPYING for details.
*/

/*  
    how to compile: 
    gcc -O2 -Wall -g -o chunky_encode chunky_encode.c 

 */


#include <stdio.h>
#include <assert.h>
#include <malloc.h>

#define Int32 int

#define dotimes(var,lim) for (var=0;var<(lim);var++) 
#define MAX(a,b) (((a)>(b)) ? (a) : (b))
#define MIN(a,b) (((a)<(b)) ? (a) : (b))
#define ABS(zz) ((zz)<0 ? -1*(zz) : (zz))

#define MAX_CHAR 85

void reverse_insertion_sort_ii (int * a, int * b, int n);

void create_huffman_code (int * sym, int * count, int n_syms,
			  unsigned Int32 * code_bits, 
			  int * code_lengths,
			  unsigned char * decoding_table);


typedef struct huff_node
{
  struct huff_node * parent;
  int which_child;
  int count;
  int leafp;

  int seq;

  struct huff_node * left;
  struct huff_node * right;

  int sym;

} hufNode;



int global_counter;

void depth_first_set_seq (hufNode * x)
{
  x->seq = global_counter++;
  if (x->leafp) return;
  depth_first_set_seq (x->left);
  depth_first_set_seq (x->right);
}


/*
            *   7
  0 1 2 3 4 5 6

  */


void create_huffman_code (int * sym, int * count, int n_syms,
			  unsigned Int32 * code_bits, 
			  int  * code_lengths,
			  unsigned char * decoding_table)
{
  hufNode * hn_store = (hufNode *) malloc ((2*n_syms-1)*sizeof(hufNode));
  hufNode ** work    = (hufNode **) malloc (n_syms*sizeof(hufNode *));
  hufNode * this, *a, *b, *new;
  int next_free, n_left, slotno;
  int i,j;

  reverse_insertion_sort_ii (count, sym, n_syms);

  dotimes (i, n_syms) {
    this = hn_store + i;
    this->parent = NULL;
    this->leafp = 1;
    this->sym = sym[i];
    this->count = count[i];
    work[i] = this;
  }
  n_left = n_syms;
  next_free = n_syms;

  while (n_left > 1) {

    a = work[n_left-2];
    b = work[n_left-1];
    new = &(hn_store[next_free++]);
    new->parent = NULL;
    new->leafp = 0;
    new->count = a->count + b->count;
    new->left = a;
    new->right = b;
    a->parent = new;
    b->parent = new;
    a->which_child = 0;
    b->which_child = 1;

    slotno = n_left - 2;

    while (slotno-1 >= 0 && work[slotno-1]->count < new->count)
      slotno--;

    for (i = n_left-3; i >= slotno; i--)
      work[i+1] = work[i];

    work[slotno] = new;
    n_left -= 1;
  }

  global_counter = 0;
  depth_first_set_seq (work[0]);

  assert (next_free == 2*n_syms-1);

  dotimes (i, 2*n_syms-1) {
    this = hn_store + i;
    if (this->leafp == 1)
      decoding_table[this->seq] = this->sym + 170;
    else
      decoding_table[this->seq] = this->right->seq;
  }

  dotimes (i, MAX_CHAR) {
    code_bits[i] = 0;
    code_lengths[i] = 0;
  }

  dotimes (i, n_syms) {
    this = hn_store + i;
    assert(this->leafp == 1);
    j = this->sym;
    while (this->parent != NULL) {
      code_lengths[j] += 1;
      code_bits[j] = (code_bits[j] << 1) | this->which_child;
      this = this->parent;
    }
  }

  free (hn_store);
  free (work);
}


int massage_sym_counts (int * sym, int * count, int n_syms)
{
  int i;
  reverse_insertion_sort_ii (count, sym, n_syms);
  i = 0;
  while (count[i] > 0) i++;
  return i;
}


void reverse_insertion_sort_ii (int * a, int * b, int n)
{
  int i,j,t,s;
  int mn, mnpos;
  dotimes (i, n-1) {
    mn = a[n-1]; mnpos = n-1;
    for (j=n-2; j >= i; j--)
      if (a[j] > mn) {
	mn = a[j]; mnpos = j;
      }
    t = a[i]; a[i] = a[mnpos]; a[mnpos] = t;
    s = b[i]; b[i] = b[mnpos]; b[mnpos] = s;
  }
  dotimes (i, n-1)
    assert (a[i] >= a[i+1]);
}


int n_words;

char ** raw_word;
char ** stripped_word;
int * stripped_word_length;


#define WORD_BUF_LEN 100

void read_words (char * file_name)
{
  FILE * infile;
  char inbuf [WORD_BUF_LEN];
  int i,j;
  infile = fopen(file_name, "r");
  assert(infile != NULL);
  n_words = 0;
  while (fgets(inbuf, WORD_BUF_LEN, infile) != NULL)
    n_words += 1;
  /* really should check for and discard empty lines */
  fclose(infile);
  fprintf(stderr, "there are %d lines in %s\n", n_words, file_name);
  infile = fopen(file_name, "r");  
  raw_word = (char **) malloc(n_words * sizeof(char *));
  stripped_word = (char **) malloc(n_words * sizeof(char *));
  stripped_word_length = (int *) malloc(n_words * sizeof(int));
  assert(raw_word != NULL);
  assert(stripped_word != NULL);
  i = 0;
  while (fgets(inbuf, WORD_BUF_LEN, infile) != NULL) {
    raw_word[i] = (char *) malloc ((strlen(inbuf)+1) * sizeof(char));
    stripped_word[i] = (char *) malloc ((strlen(inbuf)+1) * sizeof(char));
    j = 0;
    while (inbuf[j] > 25) {
      raw_word[i][j] = inbuf[j];
      j += 1;
    }
    raw_word[i][j] = 0;
    i += 1;
  }
  assert (i == n_words);
  fclose(infile);
}

void create_stripped_words ()
{
  int this_len, last_len;
  int i, j, k;
  int hack_flag;
  int n_matches;
  int end;

  stripped_word[0][0] = 0;
  stripped_word_length[0] = 1 + strlen(raw_word[0]);
  strcpy (stripped_word[0]+1, raw_word[0]);

  for (j = 1; j < n_words; j++) {
    this_len = strlen(raw_word[j]);
    last_len = strlen(raw_word[j-1]);
    end = MIN(this_len, last_len);

    i = 0;

    while (i < end) {
      if (raw_word[j][i] != raw_word[j-1][i] || i == 15) break; 
      i += 1;
    }
    n_matches = i;

    stripped_word_length[j] = 1 + this_len - n_matches;

    if (i < end && raw_word[j][i] == raw_word[j-1][i] + 1) hack_flag = 1;
    else hack_flag = 0;
    
    stripped_word[j][0] = n_matches;
    k = 1;
    while (i < this_len)
      stripped_word[j][k++] = raw_word[j][i++];
    
    if (hack_flag)
      stripped_word[j][1] = 25;
  }
}


void write_stripped_words_for_testing ()
{
  int i,j;
  dotimes (i, n_words) {
    dotimes (j, stripped_word_length[i])
      printf("%c", stripped_word[i][j]);
  }
      printf("%c", 0);
}








#define CHUNK_SIZE 5000


int int_divide_rounding_up (int x, int y)
{
  int z = x / y;
  if (z * y < x) z += 1;
  return z;
}



int n_chunks;

unsigned char ** in_chunks;
short            * in_chunk_sizes;

unsigned char ** out_chunks;
short            * out_chunk_sizes;

int * chunk_first_word_index;
int * chunk_last_word_index;

void write_chunks_for_testing ()
{
  int i,j;
  dotimes (i, n_chunks) {
    dotimes (j, in_chunk_sizes[i]-1) {
      printf("%c", in_chunks[i][j]); 
    }
    /*    printf("%d %d\n", 
	   in_chunk_sizes[i],
	   in_chunks[i][in_chunk_sizes[i]-1]);
	   */
  }
  printf("%c", 0);
}



void create_chunks ()
{
  int i,k;
  int wordno, chunkno;
  int possible_n_chunks;
  int start_of_chunk;

  int total_stripped_size = 0;

  dotimes (i, n_words) 
    total_stripped_size += stripped_word_length[i];

  possible_n_chunks = 1 + int_divide_rounding_up (total_stripped_size,
						  CHUNK_SIZE);

  chunk_first_word_index = (int *)
    malloc (possible_n_chunks * sizeof(int));

  chunk_last_word_index = (int *)
    malloc (possible_n_chunks * sizeof(int));

  in_chunk_sizes =
    (short *) malloc (possible_n_chunks * sizeof(short));

  out_chunk_sizes =
    (short *) malloc (possible_n_chunks * sizeof(short));

  in_chunks =
    (unsigned char **) malloc (possible_n_chunks * sizeof(unsigned char *));

  out_chunks =
    (unsigned char **) malloc (possible_n_chunks * sizeof(unsigned char *));

  dotimes (i, possible_n_chunks) {
    in_chunks [i] = (unsigned char *) malloc (CHUNK_SIZE);
    out_chunks[i] = (unsigned char *) malloc (CHUNK_SIZE);
  }

  wordno = 0;
  chunkno = 0;
  start_of_chunk = 1;
  k = 0;

  while (wordno < n_words) {

    if (start_of_chunk) {
      k = 0;
      in_chunks[chunkno][k++] = '\0';
      dotimes (i, strlen(raw_word[wordno]))
	in_chunks[chunkno][k++] = raw_word[wordno][i];
      chunk_first_word_index[chunkno] = wordno;
      wordno += 1;
      start_of_chunk = 0;
    }
    
    else if (k + stripped_word_length [wordno] + 1 > CHUNK_SIZE) {
      in_chunks[chunkno][k++] = '\0';      
      in_chunk_sizes[chunkno] = k;
      chunkno += 1;
      start_of_chunk = 1;
    }

    else {
      dotimes (i, stripped_word_length [wordno])
	in_chunks[chunkno][k++] = stripped_word[wordno][i];
      chunk_last_word_index[chunkno] = wordno;
      wordno += 1;
    }
  }
  if (!start_of_chunk) {
      in_chunks[chunkno][k++] = '\0';      
      in_chunk_sizes[chunkno] = k;
      chunkno += 1;
  }
  assert (chunkno <= possible_n_chunks);
  n_chunks = chunkno;

  /*
  dotimes (i, n_chunks) {
    fprintf(stderr, "%d %s %s\n", i, 
	    raw_word[chunk_first_word_index[i]], 
	    raw_word[chunk_last_word_index[i]]);
  }
  */

}





/********************************************************/

/* recode characters to lie in range 0-84 */

int remap_table [256];
int unmap_table [85];
int n_uniq = 0;


void remap_chunk_buffers ()
{
  int i,j;
  int c;

  dotimes (i, 256) remap_table[i] = -1;
  dotimes (i, 85)  unmap_table[i] = -1;

  dotimes (i, n_chunks) {
    dotimes (j, in_chunk_sizes[i]) {
      c = in_chunks[i][j];
      if (remap_table[c] == -1) {
	assert(n_uniq < 85);
	remap_table[c] = n_uniq;
	unmap_table[n_uniq] = c;
	n_uniq += 1;
      }
    }
  }
  dotimes (i, n_chunks) 
    dotimes (j, in_chunk_sizes[i]) 
    in_chunks[i][j] = remap_table[in_chunks[i][j]];

  fprintf(stderr, "there are %d unique characters after prefix stripping\n",
	  n_uniq);

}





/********************************************************/


void get_word (unsigned char * word_buf, 
	       unsigned char * data_buf,
	       int * pos)
{
  int c;
  unsigned char * bpt;

  c = unmap_table[data_buf[*pos]];
  *pos += 1;
  assert (c <= 15);
  bpt = word_buf + c;

  if (unmap_table[data_buf[*pos]] == 25) {
    *pos += 1;
    *bpt += 1;
    bpt += 1;
  }

  while ((c = unmap_table[data_buf[*pos]]) > 15) {
    *bpt++ = c;
    *pos += 1;
  }

  *bpt = '\0';
}


/********************************************************/



void write_chunk_index_file (char * name)
{
  FILE * f = fopen(name, "w");
  int i;
  assert (f != NULL);
  dotimes (i, n_chunks)
    fprintf(f, "%c%c",
	    raw_word[chunk_first_word_index[i]][0], 
	    raw_word[chunk_last_word_index[i]][0]);
  fclose (f);
}




/********************************************************/



int buf_in_use = 0;
int buf_out_bit_count = 0;
int buf_out_bit_buf = 0;
int buf_out_pos = 0;
int buf_out_len;
unsigned char * buf_out_buf;

void buf_output_initialize (unsigned char * buf, int buflen)
{
  assert (!buf_in_use);
  buf_out_bit_count = 0;
  buf_out_bit_buf = 0;  
  buf_out_pos = 0;
  buf_out_len = buflen;
  buf_out_buf = buf;
  buf_in_use = 1;
}

void buf_put_bit (int bit)
{
  assert (buf_in_use);
  if (buf_out_bit_count == 8) {
    assert (buf_out_pos < buf_out_len);
    buf_out_buf [buf_out_pos++] = buf_out_bit_buf;
    buf_out_bit_count = 0;
    buf_out_bit_buf = 0;
  }
  buf_out_bit_buf = (buf_out_bit_buf | (256 * bit)) >> 1;
  buf_out_bit_count += 1;
}


int buf_output_finalize ()
{
  assert (buf_in_use);
  if (buf_out_bit_count != 0) {
    buf_out_bit_buf = (buf_out_bit_buf >> (8 - buf_out_bit_count));
    assert (buf_out_pos < buf_out_len);
    buf_out_buf [buf_out_pos++] = buf_out_bit_buf;
  }
  buf_in_use = 0;
  return (buf_out_pos);
}






/************************************************************/


typedef struct model
{
  int in_syms   [MAX_CHAR];
  int in_counts [MAX_CHAR];

  int n_syms_occurring;

  unsigned Int32 encoder_bits [MAX_CHAR];
  int            encoder_lens [MAX_CHAR];

  unsigned char decoder [2*MAX_CHAR];

} Model;


void zero_model (Model * m)
{
  int i;
  dotimes (i, MAX_CHAR) {
    m->in_counts [i] = 0;
    m->in_syms   [i] = i;
  }
}


void model_count_char (Model * m, int c)
{
  m->in_counts[c] += 1;
}


void model_create_encoder (Model * m)
{
  m->n_syms_occurring = massage_sym_counts (m->in_syms, m->in_counts,
					    MAX_CHAR);
  if (m->n_syms_occurring == 0) return;

  /*fprintf(stderr,"%d symbols occur in this model\n",m->n_syms_occurring);*/

  create_huffman_code (m->in_syms, m->in_counts, m->n_syms_occurring,
		       m->encoder_bits, m->encoder_lens,
		       m->decoder);
}


Model glo_model;

Model cxt_model [MAX_CHAR];


int use_it [MAX_CHAR];


/*********************************************************************/


int decoder_offsets [MAX_CHAR];


void write_code_file (char * name)
{
  FILE * f = fopen(name, "w");
  int cur_offset = 2*glo_model.n_syms_occurring-1;
  int i,j;
  assert (f != NULL);

  dotimes (i, MAX_CHAR) {
    if (use_it[i]) {
      decoder_offsets [i] = cur_offset;
      cur_offset += 2*cxt_model[i].n_syms_occurring-1;
    }
    else
      decoder_offsets [i] = 0;
  }

  fprintf (f,"%d\n", n_uniq);

  dotimes (i, n_uniq)
    fprintf(f,"%d\n", unmap_table[i]);

  dotimes (i, n_uniq) {
    fprintf(f,"%d\n", decoder_offsets[i]);
  }

  fprintf (f,"%d\n", cur_offset);

  dotimes (i, 2*glo_model.n_syms_occurring-1)
    fprintf(f,"%d\n", glo_model.decoder[i]);

  dotimes (i, n_uniq) {
    if (use_it[i]) {
      dotimes (j, 2*cxt_model[i].n_syms_occurring-1)
	fprintf(f,"%d\n", cxt_model[i].decoder[j]);
    }
  }
  fclose (f);
}


void encode_and_output_symbol (Model * m, unsigned char c)
{
  int bits = m->encoder_bits [(int) c];
  int len  = m->encoder_lens [(int) c];

  while (len-- > 0) {
    buf_put_bit (bits & 1);
    bits = bits >> 1;
  }
}


/*********************************************************************/

/* The number of bits needed to encode the symbols
   summarized by the counts in c using model m
   Not all combinations of c and m are valid.
   */

int evaluate_model (Model * c, Model * m)
{
  int i;
  int total = 0;
  dotimes (i, c->n_syms_occurring)
    total += c->in_counts[i] * m->encoder_lens[c->in_syms[i]];
  return total;
}


/*********************************************************************/





int main (int argc, char **argv)
{
  int i,j,context;

  if (argc != 4) {
    fprintf(stderr, "usage: %s infile code-file chunk-index-file > bits-out\n",
	    argv[0]);
    exit(-1);
  }

  read_words (argv[1]);

  create_stripped_words ();

  /*  write_stripped_words_for_testing (); */

  create_chunks ();

  /*  write_chunks_for_testing (); */

  write_chunk_index_file (argv[3]);

  remap_chunk_buffers ();

  zero_model (&glo_model);

  dotimes (i, n_chunks) 
    dotimes (j, in_chunk_sizes[i])
    model_count_char (&glo_model, (int) in_chunks[i][j]);
  model_create_encoder (&glo_model);

  dotimes (i, MAX_CHAR) zero_model (cxt_model+i);


  dotimes (i, n_chunks) 
    dotimes (j, in_chunk_sizes[i]-1)
    model_count_char (cxt_model + (int) in_chunks[i][j],
		      (int) in_chunks[i][j+1]);
  dotimes (i, MAX_CHAR) model_create_encoder (cxt_model + i);

  {
    int cxt_bits;
    int cxt_table_size;
    int glo_bits;
    dotimes (i, MAX_CHAR) {
      if (cxt_model[i].n_syms_occurring <= 0) {
	use_it [i] = 0;
	continue;
      }
      cxt_table_size = 2 * cxt_model[i].n_syms_occurring - 1;
      cxt_bits = evaluate_model (cxt_model+i, cxt_model+i);
      glo_bits = evaluate_model (cxt_model+i, &glo_model);
      /*      fprintf(stderr, "%d %c  glo: %d,  cxt: %d = %d + %d\n",
	      i, unmap_table[i],
	      glo_bits,
	      cxt_table_size * 8 + cxt_bits,
	      cxt_table_size * 8,
	      cxt_bits); */
      if (cxt_table_size * 8 + cxt_bits < glo_bits)
	use_it [i] = 1;	
      else 
	use_it [i] = 0;
    }
  }

  write_code_file (argv[2]);

  dotimes (i, n_chunks) {
    buf_output_initialize (out_chunks[i], CHUNK_SIZE);
    context = -1;
    dotimes (j, in_chunk_sizes[i]) {
      Model * m = (((context == -1) | !use_it [context]) ?
		   &glo_model :
		   cxt_model + context);
      encode_and_output_symbol (m, in_chunks[i][j]);
      context = in_chunks[i][j];
    }
    out_chunk_sizes[i] = buf_output_finalize ();
  }


  dotimes (i, n_chunks) {
    short tmp = 1 + chunk_last_word_index[i] - chunk_first_word_index[i];
    fwrite (out_chunk_sizes+i, sizeof(short), 1, stdout);
    fwrite (in_chunk_sizes+i,  sizeof(short), 1, stdout);
    fwrite (&tmp, sizeof(short), 1, stdout);
    fwrite (out_chunks[i], 1, (int) out_chunk_sizes[i], stdout);
  }

  fprintf(stderr, "there are %d chunks\n", n_chunks);

  return (0);
}
