
/* 
   Copyright 1997, Kevin J. Lang 

   install-wordlist is free software, licensed under the 
   GNU Public License V2.

   See the file COPYING for details.

   install-wordlist is based on the copylefted pilot-link.0.3.2 package.
   You can get the original source code for that package from 
   ftp://ns1.pfnet.com/pub/PalmOS/

*/


/* how to compile install-wordlist */

/* gcc -o install-wordlist install-wordlist.c libpisock.a */ 


#include <stdio.h>
#include <stdlib.h>
#include "pi-socket.h"
#include "dlp.h"
#include "assert.h"
#include "errno.h"

#define dotimes(var,lim) for (var=0;var<(lim);var++) 
#define MAX(a,b) (((a)>(b)) ? (a) : (b))
#define MIN(a,b) (((a)<(b)) ? (a) : (b))
#define ABS(zz) ((zz)<0 ? -1*(zz) : (zz))


/**************************************************/

int n_words;


#define CHUNK_SIZE 5000


int n_chunks;

unsigned char ** chunks;

short * actual_chunk_sizes;
short * expanded_chunk_sizes;
short * chunk_word_counts;



void read_chunk_file (char * name)
{
  FILE * input_file;
  int i;

  input_file = fopen (name, "r");
  assert(input_file != NULL);

  chunks = (unsigned char **) malloc (n_chunks * sizeof(unsigned char *));

  actual_chunk_sizes   = (short *) malloc (n_chunks * sizeof(short));
  expanded_chunk_sizes = (short *) malloc (n_chunks * sizeof(short));
  chunk_word_counts    = (short *) malloc (n_chunks * sizeof(short));

  dotimes (i, n_chunks) {
    chunks[i] = (unsigned char *) malloc (CHUNK_SIZE);
    fread (actual_chunk_sizes+i, sizeof(short), 1, input_file);
    fread (expanded_chunk_sizes+i, sizeof(short), 1, input_file);
    fread (chunk_word_counts+i, sizeof(short), 1, input_file);
    fread (chunks[i], 1, actual_chunk_sizes[i], input_file);
    printf("%d %d %d\n", i, actual_chunk_sizes[i], expanded_chunk_sizes[i]);
  }

  fclose (input_file);
}




/*************************************************/

int n_uniq = 0;

unsigned char * unmap_table;

short * decoder_offsets;

unsigned char * decoder_table;

int decoder_table_len;

void read_code_file (char * name)
{
  FILE * input_file;
  int i;
  int itmp;

  input_file = fopen (name, "r");
  assert(input_file != NULL);

  fscanf(input_file, "%d", &n_uniq);
  unmap_table = (unsigned char *) malloc (n_uniq * sizeof(unsigned char));

  dotimes (i, n_uniq) {
    fscanf(input_file, "%d", &itmp);
    unmap_table [i] = itmp;
  }

  decoder_offsets = (short *) malloc (n_uniq * sizeof(short));

  dotimes (i, n_uniq) {
    fscanf(input_file, "%d", &itmp);
    decoder_offsets [i] = (short) itmp;
  }

  fprintf(stderr, "decoder_offsets[3] = %d\n", (int) decoder_offsets[3]);

  fscanf(input_file, "%d", &decoder_table_len);
  decoder_table = (unsigned char *) malloc (decoder_table_len
					    * sizeof(unsigned char));

  dotimes (i, decoder_table_len) {
    fscanf(input_file, "%d", &itmp);
    decoder_table [i] = itmp;
  }

  fclose (input_file);

}


/**************************************************/


int index_len;
unsigned char * index_string;


void read_index_file (char * name)
{
  FILE * input_file;
  int ret;

  input_file = fopen (name, "r");
  assert(input_file != NULL);

  fseek(input_file, 0, SEEK_END);
  index_len = ftell(input_file);
  fseek(input_file, 0, SEEK_SET);

  index_string = (unsigned char *) malloc (index_len);
  assert(index_string != NULL);

  ret = fread (index_string, 1, index_len, input_file);
  assert (ret == index_len);

  fclose (input_file);
}



/**************************************************/


typedef struct header_rec
{
  short n_words;
  short n_chunks;
} headerRec;

char * header_buf;
int header_buf_len;

void pack_header_buf ()
{
  set_short (header_buf+0*sizeof(short), n_chunks);
  set_long  (header_buf+1*sizeof(short), n_words);
}


/**************************************************/

char * offset_buf;
int offset_buf_len;

void pack_offset_buf ()
{
  int i;
  dotimes (i, n_uniq)
    set_short (offset_buf+i*sizeof(short), decoder_offsets[i]);
}

/**************************************************/



typedef struct chunk_rec
{
  short chunk_len;
  unsigned char chunk_contents [CHUNK_SIZE];
} chunkRec;


char * chunk_buf;
int chunk_buf_len;

void pack_chunk_buf (int i)
{
  set_short (chunk_buf+0*sizeof(short), (short) expanded_chunk_sizes [i]);
  set_short (chunk_buf+1*sizeof(short), (short) chunk_word_counts [i]);
  memcpy   ((void *) (chunk_buf+2*sizeof(short)),
	    (void *) chunks [i],
	    actual_chunk_sizes [i]);
}






/**************************************************/



void create_pilot_buffers ()
{
  header_buf_len = 1 * sizeof(short) + sizeof(long);
  header_buf     = (char *) malloc (header_buf_len);

  chunk_buf_len = 2*sizeof(short) + CHUNK_SIZE;
  chunk_buf     = (char *) malloc (chunk_buf_len);

  offset_buf_len = n_uniq * sizeof(short);
  offset_buf     = (char *) malloc (offset_buf_len);

  pack_header_buf();
  pack_offset_buf();
}





/**************************************************/




int main(int argc, char *argv[])
{
  struct pi_sockaddr addr;
  int db;
  int sd;
  struct PilotUser U;
  int ret;
  int status;
  int i;

  if (argc != 8) {
    fprintf(stderr,
	    "usage: %s %s bit-file code-file n-words n-chunks index-file db-name\n",
	    argv[0],
	    TTYPrompt);
    exit(2);
  }

  n_words  = atoi  (argv[4]);
  n_chunks = atoi  (argv[5]);

  read_chunk_file  (argv[2]);
  read_code_file   (argv[3]);
  read_index_file  (argv[6]);

  create_pilot_buffers();

  if (!(sd = pi_socket(AF_SLP, SOCK_STREAM, PF_PADP))) {
    perror("pi_socket");
    exit(1);
  }
    
  addr.sa_family = AF_SLP;
  addr.port = 3;
  strcpy(addr.device,argv[1]);
  
  ret = pi_bind(sd, &addr, sizeof(addr));
  if(ret == -1) {
    perror("pi_bind");
    exit(1);
  }

  ret = pi_listen(sd,1);
  if(ret == -1) {
    perror("pi_listen");
    exit(1);
  }

  sd = pi_accept(sd, 0, 0);
  if(sd == -1) {
    perror("pi_accept");
    exit(1);
  }

  /* Ask the pilot who it is. */
  dlp_ReadUserInfo(sd,&U);
  
  /* Tell user (via Pilot) that we are starting things up */
  dlp_OpenConduit(sd);
  



  /*
  status = dlp_CreateDB(sd, (long) CREATOR, (long) 1, 0, 0, 0,
			argv[7],
			&discard);
  if (status < 0) {
    puts("Unable to create database");
    printf("return val = %d, errno = %d\n", status, errno);
    dlp_AddSyncLogEntry(sd, "Unable to create database.\n");
    exit(1);
  }
  */


  status = dlp_OpenDB(sd, 0, 0x80|0x40, argv[7], &db);

  if (status < 0) {
    puts("Unable to open database");
    printf("return val = %d, errno = %d\n", status, errno);
    dlp_AddSyncLogEntry(sd, "Unable to open database.\n");
    exit(1);
  }

  dlp_WriteRecord(sd,(unsigned char)db,0,0,0, header_buf, header_buf_len, 0);

  dlp_WriteRecord(sd,(unsigned char)db,0,0,0,
		  unmap_table,
		  n_uniq,
		  0);

  dlp_WriteRecord(sd,(unsigned char)db,0,0,0, offset_buf, offset_buf_len, 0);

  dlp_WriteRecord(sd,(unsigned char)db,0,0,0,
		  decoder_table, 
		  decoder_table_len,
		  0);

  dlp_WriteRecord(sd,(unsigned char)db,0,0,0,
		  index_string,
		  index_len,
		  0);

  dotimes (i, n_chunks) {
    pack_chunk_buf (i);
    dlp_WriteRecord(sd,(unsigned char)db,0,0,0,
		    chunk_buf,
		    actual_chunk_sizes[i] + 2*sizeof(short),
		    0);
  }

  /* Close the database */
  dlp_CloseDB(sd, db);

  /* Tell the user who it is, with a different PC id. */
  U.lastSyncPC = 0x00010000;
  U.succSyncDate = time(NULL);
  U.lastSyncDate = U.succSyncDate;
  dlp_WriteUserInfo(sd,&U);

  dlp_AddSyncLogEntry(sd, "Wrote wordlist to Pilot.\n");
  
  /* All of the following code is now unnecessary, but harmless */
  

  dlp_EndOfSync(sd,0);
  pi_close(sd);
  exit(0);
}
